"""
preprocessing and shorthand functions for dataframes, mappings, and ndarrays.
"""
from functools import reduce
from itertools import product
from operator import attrgetter
import re
from typing import Any, Optional

from cytoolz import valfilter
import numpy as np
import pandas as pd

from dustgoggles.structures import listify


def itemize_numpy(obj: Any):
    """
    convert objects of numpy dtypes to python scalars. in this context,
    primarily for json serialization.
    """
    if isinstance(obj, np.generic):
        return obj.item()
    return obj


def dupe_df_block(dataframe: pd.DataFrame, repeats: int):
    """
    create a new dataframe by duplicating "dataframe" "repeats" times
    """
    return pd.DataFrame(
        np.repeat(dataframe.values, repeats, axis=0),
        columns=dataframe.columns,
    )


# TODO, maybe: allow this to work with columns that have collections
#  in individual cells
def check_and_drop_duplicate_columns(dataframe: pd.DataFrame):
    """
    drop duplicate columns -- in both name and value -- from a dataframe.
    if equal-named columns are _not_ equal-valued, raises a ValueError.
    """
    extra_columns = dataframe.columns[dataframe.columns.duplicated()]
    if len(extra_columns) == 0:
        return dataframe
    for column in extra_columns:
        test_equality = dataframe.loc[:, column] == dataframe.loc[:, column]
        if not test_equality.all(axis=None):
            raise ValueError
    return dataframe.loc[:, ~dataframe.columns.duplicated()]


def extract_constants(
    df: pd.DataFrame,
    to_dict=True,
    drop_constants=False,
    how="rows",
    dropna=True
):
    """
    extract 'constant' values from a dataframe -- by default, columns with
    the same value in each row; if how == 'columns', then indices with the
    same value in each column. if to_dict is True, transforms them into a
    dictionary with keys equal to column/index names and values equal to the
    constant values. if drop_constants is True, also removes these
    constant-valued rows/columns from the returned dataframe.
    returns the constants (as a dict if specified) and either the
    variable-only or full original dataframe (as specified)
    """
    if how not in ("rows", "columns"):
        raise ValueError(f"unknown how {how}")
    if how == "rows":
        constant_indices = pd.Series(False, df.columns)
    else:
        constant_indices = pd.Series(False, df.index)
    try:
        method = "iterrows" if how == "columns" else "items"
        for ix, series in getattr(df, method)():
            series = series if dropna is False else series.dropna()
            if (series == series.iloc[0]).all():
                constant_indices.loc[ix] = True
    except TypeError:
        # TODO: still necessary?
        if how == "columns":
            raise NotImplementedError("nested fields not supported columnwise")
        for c in df.columns:
            if isinstance(df[c].iloc[0], list):
                test_series = df[c].map(tuple)
            else:
                test_series = df[c]
            if (test_series.iloc[0] == test_series).all():
                constant_indices.loc[c] = True
    if how == "rows":
        constants = df.loc[:, constant_indices]
        variables = df.loc[:, ~constant_indices]
    else:
        constants = df.loc[constant_indices]
        variables = df.loc[~constant_indices]
    if to_dict:
        constants = constants.iloc[0].to_dict()
    if drop_constants:
        return constants, variables
    return constants, df


def split_on(
    df: pd.DataFrame, predicate: pd.Series
) -> [pd.DataFrame, pd.DataFrame]:
    return df.loc[predicate], df.loc[~predicate]


def pdstr(str_method_name, *str_args, **str_kwargs):
    """
    creates a mappable function that accesses .str methods of passed Series
    """

    def replacer(series: pd.Series):
        method = getattr(series.str, str_method_name)
        return method(*str_args, **str_kwargs)

    return replacer


def typed_columns(df, typenames):
    """return columns of df of types that match patterns given in typenames."""
    if isinstance(typenames, str):
        typenames = (typenames,)
    if isinstance(df.dtypes, np.dtype):
        if re.match("|".join(typenames).lower(), df.dtypes.name.lower()):
            return df
        return pd.DataFrame(index=df.index)
    names = df.dtypes.apply(attrgetter("name")).str.lower()
    type_pattern = f"u?({'|'.join(typenames)})".lower()
    return df[df.dtypes.loc[names.str.match(type_pattern)].index]


def numeric_columns(df):
    """return 'numeric" columns of df."""
    return typed_columns(df, ("int", "float"))


DTYPES = (
    "uint8", "int8", "uint16", "int16",
    "uint32", "int32", "uint64", "int64",
    "float32", "float64"
)


def downcast(arr, atol=0.01, rtol=0.001):
    # NaNs and infs don't count but should
    # be preserved
    nonfinite = ~np.isfinite(arr)
    nonfinite_vals = arr[nonfinite]
    arr[nonfinite] = 0
    recast = None
    for dtype in DTYPES:
        recast = arr.astype(dtype)
        offsets = np.abs(arr - recast)
        if (aerr := np.nanmax(np.abs(offsets))) > atol:
            continue
        if (rerr := np.nanmax(np.abs(offsets / arr))) > rtol:
            continue
        break
    if recast is None:
        raise TypeError
    # noinspection PyUnboundLocalVariable
    return {
        'recast': recast,
        'nonfinite_mask': nonfinite,
        'nonfinite_vals': nonfinite_vals,
        'aerr': aerr,
        'rerr': rerr
    }


def downcast_df(df, atol=0.01, rtol=0.001):
    num = numeric_columns(df)
    cast_series, cast_records = [], []
    for name, col in num.items():
        rec = downcast(col.values.T.copy(), atol, rtol) | {'name': name}
        cast_records.append(rec)
    while len(cast_records) > 0:
        rec = cast_records.pop()
        hasna = rec['nonfinite_mask'].any()
        if hasna and ("int" in (dname := rec['recast'].dtype.name)):
            prefix = "U" if dname.startswith("u") else ""
            depth = (rec['recast'].dtype.itemsize * 8)
            dtype = getattr(pd, f"{prefix}Int{depth}Dtype")()
            series = pd.Series(rec['recast'], name=rec['name'], dtype=dtype)
        else:
            series = pd.Series(rec['recast'], name=rec['name'])
        if hasna:
            series.loc[rec['nonfinite_mask']] = rec['nonfinite_vals']
        cast_series.append(series)
    castdown = pd.concat(list(reversed(cast_series)), axis=1)
    castdown.index = df.index
    # TODO: inefficient inserts
    for c in castdown:
        df[c] = castdown[c]
    return df


def junction(df1, df2, columns, set_method="difference"):
    keys1 = df1[list(columns)].value_counts().index.to_list()
    keys2 = df2[list(columns)].value_counts().index.to_list()
    return getattr(set(keys1), set_method)(set(keys2))


def smash(df, by, values=None):
    by = listify(by)
    if values is not None:
        df = df.loc[df[by].isin(values).all(axis=1)]
    df = df.melt(by)
    names = reduce(
        lambda x, y: x + "_" + y,
        [v for _, v in df[by + ["variable"]].astype(str).items()],
    )
    return pd.Series(df["value"].to_numpy(), index=names)


def unpack(series: pd.Series):
    """unpack nested listlikes to columns"""
    dropped = series.dropna()
    df = pd.DataFrame(dropped.tolist(), index=dropped.index)
    df.columns = [f"{dropped.name}_{i}" for i, _ in enumerate(df.columns)]
    return df


def unpack_column(df: pd.DataFrame, colname):
    """unpack a nested listlike column by name"""
    unpacked = unpack(df[colname])
    df.loc[unpacked.index, unpacked.columns] = unpacked
    return df.drop(columns=colname)


def unique_to_records(df, cols):
    unique_tuples = df[cols].value_counts().index.to_list()
    records = [
        {col: value for col, value in zip(cols, values)}
        for values in unique_tuples
    ]
    # noinspection PyTypeChecker
    return pd.DataFrame.from_dict(records)


def squeeze(series: pd.Series, aggfunc: Optional[str] = None) -> pd.Series:
    """
    removes duplicate index labels from a Series, aggregating values
    associated with each duplicate label using `aggfunc`. if aggfunc is
    None, `squeeze` uses "any" if the Series is boolean and simply returns
    the first element associated with each unique index label if the Series
    is not.
    """
    if aggfunc is None:
        if series.dtype.char == "?":
            aggfunc = "any"
        else:
            return series.loc[~series.index.duplicated()]
    df = pd.DataFrame(series).reset_index()
    ixname = "index" if series.index.name is None else series.index.name
    sname = 0 if series.name is None else series.name
    squeezed = df.pivot_table(
        values=series.name,
        index=ixname,
        aggfunc=aggfunc,
    )[sname]
    squeezed.index.name = series.index.name
    squeezed.name = series.name
    return squeezed



"""structured data and manipulators thereof"""
from collections import defaultdict, OrderedDict
from copy import copy
from functools import reduce, partial
from itertools import chain
from multiprocessing import Pool
from operator import methodcaller, add, getitem, eq
from typing import (
    Mapping,
    Collection,
    Any,
    Union,
    Sequence,
    MutableMapping,
    Callable,
    Type,
    Hashable,
)

from cytoolz import merge
from more_itertools import chunked, divide

from dustgoggles.func import naturals, is_it, splat


def to_records(nested: Mapping, accumulated_levels=None, level_names=None):
    level_names = naturals() if level_names is None else iter(level_names)
    records = []
    accumulated_levels = (
        {} if accumulated_levels is None else accumulated_levels
    )
    level_name = next(level_names)
    for category, mapping in nested.items():
        if all([isinstance(value, Mapping) for value in mapping.values()]):
            branch = accumulated_levels.copy()
            branch[level_name] = category
            records += to_records(mapping, branch, copy(level_names))
        else:
            category_dict = accumulated_levels | {level_name: category}
            flat = mapping | category_dict
            records.append(flat)

    return records


def _unnest(nested, mtypes, escape, levels):
    if not isinstance(nested, mtypes):
        return nested
    prefix = f"{escape}_{escape}".join(map(str, levels))
    if len(prefix) > 0:
        prefix = f"{escape}{prefix}{escape}_"
    long_records = []
    for level, maybe_mapping in nested.items():
        if not isinstance(maybe_mapping, mtypes):
            long_records.append({f"{prefix}{level}": maybe_mapping})
            continue
        long_records += _unnest(maybe_mapping, mtypes, escape,
                                levels + [level])
    return long_records


def unnest(nested, mtypes=(dict,), escape=""):
    flat_records = _unnest(nested, mtypes, escape, [])
    return merge(flat_records)


class NestingDict(defaultdict):
    """
    shorthand for automatically-nesting dictionary -- i.e.,
    insert a series of keys at any depth into a NestingDict
    and it automatically creates all needed levels above.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.default_factory = NestingDict

    __repr__ = dict.__repr__

    def todict(self):
        """warning: this function may permanently modify lower levels"""
        # note: could be made depth-first by using methodcaller('todict')
        # rather than dict
        return dict(
            dig_and_edit(
                self, lambda _, v: is_it(NestingDict)(v), lambda _, v: dict(v)
            )
        )


def listify(thing):
    """Always a list, for things that want lists"""
    if isinstance(thing, Collection):
        if not isinstance(thing, str):
            return list(thing)
    return [thing]


def reindex_mapping(mapping: Mapping) -> Mapping[int, Any]:
    assert isinstance(mapping, Mapping), "reindex_mapping only takes Mappings"
    return {
        ix: value for ix, value in zip(range(len(mapping)), mapping.values())
    }


def enumerate_as_mapping(sequence):
    if sequence is None:
        return {}
    if isinstance(sequence, Mapping):
        return sequence
    return {ix: element for ix, element in enumerate(sequence)}


def insert_after(
    new_key: Any, new_value: Any, prior_key: Any, mapping: Mapping
) -> Mapping:
    new_dict = {}
    for key, value in mapping.items():
        new_dict[key] = value
        if key == prior_key:
            new_dict[new_key] = new_value
    return new_dict


def get_from_all(key, mappings, default=None):
    """
    get all values of "key" from each dict or whatever in "mappings"
    """
    if isinstance(mappings, Mapping):
        view = mappings.values()
    else:
        view = mappings
    return list(map(methodcaller("get", key, default), view))


def getitemattr(collection: Union[Sequence, Mapping], key: Any) -> Any:
    """
    getter that attempts both getattr and getitem (intended
    for named tuples nested inside of dicts, etc)
    """
    try:
        return getitem(collection, key)
    except (KeyError, IndexError, TypeError):
        return getattr(collection, key)


def get_from(collection, keys, default=None):
    """
    hierarchical (toolz-style) extension of get_itemattr() --
    (hierarchical list of keys, collection ->
    item of collection, possibly from a nested collection)
    """
    level = collection
    for key in keys:
        try:
            level = getitemattr(level, key)
        except AttributeError:
            return default
    return level


def constant_to_dig_predicate(putative_predicate, match="key", default=eq):
    if isinstance(putative_predicate, Callable):
        return putative_predicate
    if match == "key":
        return keyonly(partial(default, putative_predicate))
    elif match == "value":
        return valonly(partial(default, putative_predicate))
    elif match == "item":
        return partial(default, putative_predicate)


def keyonly(func):
    def onlykey(key, _value):
        return func(key)

    return onlykey


def valonly(func):
    def onlyval(_key, value):
        return func(value)

    return onlyval


def _evaluate_diglevel(mapping, predicate, mtypes):
    # note: this is a little awkward and is done to circumvent the fact
    # that checking isinstance against the ABCs is relatively slow
    # but we would like to support many types of Mapping
    # that are not dreamt of in this library.
    nests = [v for v in mapping.values() if isinstance(v, mtypes)]
    level_items = [(k, v) for k, v in mapping.items() if predicate(k, v)]
    return level_items, nests


def dig_for_all(mapping, predicate, mtypes):
    level_items, nests = _evaluate_diglevel(mapping, predicate, mtypes)
    if nests:
        level_items += reduce(
            add, [dig_for_all(nest, predicate, mtypes) for nest in nests]
        )
    return level_items


def dig_all_wrapper(mapping, ref, match, base_pred, element_ix, mtypes):
    predicate = constant_to_dig_predicate(ref, match, base_pred)
    items = dig_for_all(mapping, predicate, mtypes)
    if len(items) == 0:
        return None
    if element_ix is None:
        return items
    return [item[element_ix] for item in items]


def dig_for_values(mapping, ref, match="key", base_pred=eq, mtypes=(dict,)):
    return dig_all_wrapper(mapping, ref, match, base_pred, 1, mtypes)


def dig_for_keys(mapping, ref, match="key", base_pred=eq, mtypes=(dict,)):
    return dig_all_wrapper(mapping, ref, match, base_pred, 0, mtypes)


def dig_for_items(mapping, ref, match="key", base_pred=eq, mtypes=(dict,)):
    return dig_all_wrapper(mapping, ref, match, base_pred, None, mtypes)


def dig_for(mapping, predicate, mtypes):
    """
    return the first item, using a breadth-first search, matching ref
    """
    level_items, nests = _evaluate_diglevel(mapping, predicate, mtypes)
    if level_items:
        return level_items
    if not nests:
        return None
    dug, iternests = None, iter(nests)
    while dug is None:
        try:
            dug = dig_for(next(iternests), predicate, mtypes)
        except StopIteration:
            return None
    return dug


def dig_wrapper(mapping, ref, match, base_pred, element_ix, mtypes):
    predicate = constant_to_dig_predicate(ref, match, base_pred)
    item = dig_for(mapping, predicate, mtypes)
    if item is None:
        return None
    if element_ix is None:
        return item[0]
    return item[0][element_ix]


def dig_for_value(mapping, ref, match="key", base_pred=eq, mtypes=(dict,)):
    return dig_wrapper(mapping, ref, match, base_pred, 1, mtypes)


def dig_for_key(mapping, ref, match="key", base_pred=eq, mtypes=(dict,)):
    return dig_wrapper(mapping, ref, match, base_pred, 0, mtypes)


def dig_for_item(mapping, ref, match="key", base_pred=eq, mtypes=(dict,)):
    return dig_wrapper(mapping, ref, match, base_pred, None, mtypes)


def dig_and_edit(
    mapping: MutableMapping,
    filter_func: Callable[[Any, Any], Any],
    setter_func: Callable[[Any, Any], Any],
    mtypes: tuple[Type[MutableMapping]] = (dict,),
) -> MutableMapping:
    matches = tuple(filter(splat(filter_func), mapping.items()))
    for key, value in matches:
        mapping[key] = setter_func(key, value)
    for nest in [v for v in mapping.values() if isinstance(v, mtypes)]:
        dig_and_edit(nest, filter_func, setter_func, mtypes)
    return mapping


# TODO: turn pivot.split_on into a dispatch function in structures,
#  replace downstream
def separate_by(collection, ref):
    hits = []
    misses = []
    for item in collection:
        if ref(item) is True:
            hits.append(item)
        else:
            misses.append(item)
    return hits, misses


def rmerge(map1, map2, mtypes=(dict, OrderedDict)):
    """
    recursively merge map1 and map2. result is similar to map1 | map2, but
    also merges any mappings at lower levels. for example:

    >>> rmerge({'a': 1, 'b': {'a': 1}}, {'b': {'b': 1}, 'c': 2})
    {'a': 1, 'b': {'a': 1, 'b': 1}, 'c': 2}
    """
    output = {}
    only_in_one = set(map1.keys()).symmetric_difference(map2.keys())
    for kv in chain(map1.items(), map2.items()):
        if kv[0] in only_in_one:
            output[kv[0]] = kv[1]
        elif all(map(lambda m: isinstance(m, mtypes), (map1[kv[0]], map2[kv[0]]))):
            output[kv[0]] = rmerge(map1[kv[0]], map2[kv[0]])
        else:
            output[kv[0]] = map2[kv[0]]
    return output


class HashDict:
    def __init__(self, equivalence: Callable[[Any], Hashable] = hash):
        self.dict_ = {}
        self.reverse = {}
        self.hasher = equivalence

    def __setitem__(self, key, item):
        itemhash = self.hasher(item)
        self.dict_[key] = itemhash
        if itemhash not in self.reverse.keys():
            self.reverse[itemhash] = item

    def __getitem__(self, key):
        return self.reverse[self.dict_[key]]

    def __str__(self):
        return (
            f"HashDict with {len(self.dict_)} keys storing "
            f"{len(self.reverse)} unique values"
        )

    def __repr__(self):
        return self.__str__()


class MaybeResult:
    def __init__(self, func, args, kwargs):
        try:
            self.value = func(*args, **kwargs)
        except KeyboardInterrupt:
            raise
        except Exception as ex:
            self.value = ex

    def get(self):
        if isinstance(self.value, Exception):
            raise self.value
        return self.value

    @staticmethod
    def ready():
        return True


def _chunkwrap(func, argrecs):
    return {
        rec['key']: func(*rec['args'], **rec['kwargs'])
        for rec in argrecs
    }


class MaybePool:
    def __init__(self, threads=None):
        self.threads, self.results, self.chunked = threads, {}, False
        if threads is None:
            self.pool = None
        else:
            self.pool = Pool(threads)

    def map(self, func, argrecs, as_chunks=False):
        for i, rec in enumerate(argrecs):
            rec['key'] = rec.get('key', i)
            rec['args'] = rec.get('args', ())
            rec['kwargs'] = rec.get('kwargs', {})
        if (as_chunks is False) or (self.pool is None):
            self.results = {
                rec['key']: self.apply(func, rec['args'], rec['kwargs'])
                for rec in argrecs
            }
        else:
            self.chunked = True
            chunks = divide(self.threads, argrecs)
            self.results = {
                f"chunk_{i}": self.apply(_chunkwrap, (func, chunk))
                for i, chunk in enumerate(chunks)
            }

    def apply(self, func, args=(), kwargs=None):
        """note: does apply_async by default"""
        kwargs = {} if kwargs is None else kwargs
        if self.pool is None:
            return MaybeResult(func, args, kwargs)
        return self.pool.apply_async(func, args, kwargs)

    def close(self):
        if self.pool is None:
            return
        return self.pool.close()

    def join(self):
        if self.pool is None:
            return
        return self.pool.join()

    def get(self, raise_exc=False):
        output = {}
        for k, v in self.results.items():
            try:
                if self.chunked is True:
                    output |= v.get()
                else:
                    output[k] = v.get()
            except KeyboardInterrupt:
                raise
            except Exception as ex:
                if raise_exc:
                    raise ex
                output[k] = ex
        return output

    def results_ready(self):
        return {k: v.ready() for k, v in self.results.items()}

    def ready(self):
        return all(v.ready() for v in self.results.values())

    def terminate(self):
        if self.pool is not None:
            self.pool.terminate()
        self.results = {}


def map_into_pool(
    func,
    argrecs,
    threads=None,
    filter_exc=False,
    as_chunks=True
):
    pool = MaybePool(threads)
    try:
        pool.map(func, argrecs, as_chunks=as_chunks)
        pool.close()
        pool.join()
        results = pool.get()
        if filter_exc is True:
            results = {
                k: v for k, v in results.items()
                if not isinstance(v, Exception)
            }
    finally:
        pool.terminate()
    return results



from itertools import product

from cytoolz import keyfilter
from more_itertools import windowed
import pyarrow as pa
import pyarrow.compute as pac
from pyarrow import parquet

from dustgoggles.pivot import extract_constants


BRIEF_PQ_META_FIELDS = (
    "physical_type",
    "path_in_schema",
    "compression",
    "encodings",
    "has_dictionary_page",
    "col_ix",
)


def parquet_metadata_records(parquet_fn):
    meta = parquet.read_metadata(parquet_fn)
    records = []
    for col_ix, group_ix in product(
        range(meta.num_columns), range(meta.num_row_groups)
    ):
        record = meta.row_group(group_ix).column(col_ix).to_dict()
        if "statistics" in record.keys():
            record |= record.pop("statistics")
        records.append(record | {"row_group": group_ix, "col_ix": col_ix})
    return records


def _flatten_meta_df(meta_df, extended):
    import pandas as pd

    col_series = []
    for col in meta_df["path_in_schema"].unique():
        col_slice = meta_df.loc[meta_df["path_in_schema"] == col]
        constants, _ = extract_constants(col_slice)
        if extended is False:
            constants = keyfilter(
                lambda k: k in BRIEF_PQ_META_FIELDS, constants
            )
        col_stats = col_slice[
            ["num_values", "total_compressed_size", "total_uncompressed_size"]
        ].sum()
        col_series.append(pd.concat([pd.Series(constants), col_stats]))
    return pd.concat(col_series, axis=1).T


def meta_record_df(meta_records, extended=False):
    import pandas as pd

    meta_df = pd.DataFrame(meta_records)
    if meta_df["path_in_schema"].duplicated().any():
        meta_df = _flatten_meta_df(meta_df, extended)
    meta_df = meta_df.convert_dtypes().rename(
        columns={
            "physical_type": "type",
            "path_in_schema": "column",
            "has_dictionary_page": "dict_page",
            "total_compressed_size": "comp_size",
            "total_uncompressed_size": "uncomp_size",
        }
    )
    if extended is False:
        meta_df = meta_df.reindex(
            columns=[
                "column",
                "type",
                "comp_size",
                "uncomp_size",
                "num_values",
                "dict_page",
                "encodings",
                "compression",
                "col_ix",
            ]
        )
    return meta_df


def meta_column_df(parquet_fn, extended=False):
    metadata = meta_record_df(
        parquet_metadata_records(parquet_fn), extended=extended
    )
    metadata.index = metadata["column"]
    metadata.index.name = None
    return metadata.T


# noinspection PyUnresolvedReferences
def read_between(
    parquet_file, field, bounds, meta_records, columns=None, verbose=True
):
    """
    managed read-between bounds operation, more efficient in some cases
    than parquet.read_table with specified filters
    """
    bound_records = []
    for rec in meta_records:
        if not ((rec["min"] > bounds[1]) or (rec["max"] < bounds[0])):
            bound_records.append(rec)
    if verbose is True:
        print(len(bound_records), len(meta_records))
    if len(bound_records) == 0:
        return
    bound_groups = []

    for rec in bound_records:
        group = parquet_file.read_row_group(rec["row_group"])
        less_than = pac.filter(group, pac.less(group[field], bounds[1]))
        more_than = pac.filter(
            less_than, pac.greater_equal(less_than[field], bounds[0])
        )
        bound_groups.append(more_than)
    return pa.concat_tables(bound_groups)


def sort_chunked(
    input_file,
    output_file,
    sort_column,
    direction="ascending",
    schema=None,
    row_group_size=None,
    use_dictionary=None,
    n_chunks=10,
    inline_processor=None,
    verbose=True,
    version="2.6"
):
    import numpy as np

    write_kwargs, open_kwargs = {}, {}
    if row_group_size is not None:
        write_kwargs["row_group_size"] = row_group_size
    if use_dictionary is not None:
        open_kwargs["use_dictionary"] = use_dictionary
    sort_reader = parquet.ParquetFile(input_file)
    meta_records = parquet_metadata_records(input_file)
    relevant_records = [
        rec for rec in meta_records if rec["path_in_schema"] == sort_column
    ]
    if schema is None:
        schema = sort_reader.read_row_group(0).schema
    sort_writer = parquet.ParquetWriter(
        output_file, version=version, schema=schema, **open_kwargs
    )
    max_value = max([rec["max"] for rec in relevant_records])
    min_value = min([rec["min"] for rec in relevant_records])
    windows = list(windowed(np.linspace(min_value, max_value, n_chunks), 2))
    windows[-1] = (windows[-1][0], windows[-1][1] + 1)
    try:
        for window in windows:
            window_table = read_between(
                sort_reader, sort_column, window, relevant_records, verbose
            )
            if window_table is None:
                continue
            if len(window_table) == 0:
                continue
            if verbose is True:
                print(window, len(window_table))
            window_table = window_table.sort_by([(sort_column, direction)])
            if inline_processor is not None:
                window_table = inline_processor(window_table)
            sort_writer.write_table(window_table, **write_kwargs)
            del window_table
    finally:
        sort_writer.close()



"""utilities for dynamic function generation and code introspection"""
from inspect import getsource, signature, Signature, getmodule
from types import FunctionType, CodeType, MethodType
from typing import Optional, Callable, MutableSequence

from cytoolz import nth

from dustgoggles.func import optionalize


# TODO: some kind of "FunctionLike" type

class AlreadyLoadedError(ValueError):
    pass


class UnreadyError(ValueError):
    pass


# TODO, maybe: optimization stuff re: kwarg mapping, etc. -- but if you wanted
#  to engage in really high call volume, you could also just explicitly call
#  the wrapped function...
def digsource(obj: Callable):
    """
    wrapper for inspect.getsource that attempts to work on objects like
    Dynamic, functools.partial, etc.
    """
    if isinstance(obj, (FunctionType, MethodType)):
        return getsource(obj)
    if "func" in dir(obj):
        # noinspection PyUnresolvedReferences
        return getsource(obj.func)
    raise TypeError(f"cannot get source for type {type(obj)}")


def get_codechild(code: CodeType, ix: int = 0) -> CodeType:
    return nth(ix, filter(lambda c: isinstance(c, CodeType), code.co_consts))


def dontcare(func: Callable, target: MutableSequence) -> Callable:
    def record_exception(exc):
        target.append(exc_report(exc) | {'func': func, 'category': 'call'})

    return optionalize(func, exc_callback=record_exception)


def compile_source(source: str):
    return get_codechild(compile(source, "", "exec"))


def define(code: CodeType, globals_: Optional[dict] = None) -> FunctionType:
    globals_ = globals_ if globals_ is not None else globals()
    return FunctionType(code, globals_)


def exc_report(exc, stringify_exception=True):
    if exc is None:
        return {}
    tb = exc.__traceback__
    exc = exc if stringify_exception is False else f"{type(exc)}: {exc}"
    report = {'exception': exc, 'lineno': [], 'name': [], 'filename': []}
    while tb is not None:
        fr = tb.tb_frame
        if (module := getmodule(fr)) is not None:
            report['name'].append(f'{module.__name__}.{fr.f_code.co_qualname}')
        else:
            report['name'].append(fr.f_code.co_name)
        report['lineno'].append(tb.tb_lineno)
        report['filename'].append(fr.f_code.co_filename)
        del fr  # taking excessive care here
        tb = tb.tb_next
    return report


class Dynamic:
    """
    simple class to help manage function definition / execution from
    dynamically-generated source.
    """
    def __init__(
        self,
        source: Optional[str] = None,
        globals_: Optional[dict] = None,
        optional: bool = False,
        lazy: bool = False,
        load_on_call: bool = True
    ):
        self.globals_ = globals_
        self.optional = optional
        self.errors = []
        self.load_on_call = load_on_call
        self.lazy = lazy
        self.call_fail = False
        self.compile_fail = False
        self.source = source
        if lazy is False:
            try:
                self.load()
            # should only encounter this when called from a class constructor
            except AlreadyLoadedError:
                pass

    # TODO: deal with modules not imported at compile-time
    #  -- maybe just permit second compilation --
    #  -- or do we need to pass globals? ugh.

    def load(self, reload=False):
        if (reload is False) and (self.func is not None):
            raise AlreadyLoadedError("self.func already loaded")
        self.compile_source(reload)
        self.define(reload)

    def compile_source(self, recompile=True):
        if (recompile is False) and (self.code is not None):
            raise AlreadyLoadedError("self.code already compiled")
        try:
            self.code = compile_source(self.source)
        except KeyboardInterrupt:
            raise
        except Exception as ex:
            self.errors.append(exc_report(ex) | {'category': 'compile'})
            self.compile_fail = True

    def define(self, redefine=False):
        if (self.func is not None) and (redefine is not True):
            raise AlreadyLoadedError("self.func already defined")
        self.func = define(self.code, self.globals_)
        self.__signature__ = signature(self.func)
        self.__name__ = self.func.__name__

    def unload(self):
        del self.code, self.func, self.errors
        self.call_fail, self.compile_fail = False, False
        self.code, self.func, self.errors = None, None, []
        self.__name__ = self.__class__.__name__
        self.__signature__ = None

    def _maybe_load_on_call(self, reload=False):
        if self.func is not None:
            return
        if self.load_on_call is True:
            return self.load(reload)
        raise UnreadyError("No loaded function.")

    def __call__(self, *args, _optional=None, **kwargs):
        self._maybe_load_on_call()
        if _optional is None:
            _optional = self.optional
        if _optional is False:
            # noinspection PyUnresolvedReferences
            return self.func(*args, **kwargs)
        try:
            return dontcare(self.func, self.errors)(*args, **kwargs)
        finally:
            if len(self.errors) > 0:
                if self.errors[-1].get('category') == 'call':
                    self.call_fail = True

    def __str__(self):
        if self.func is None:
            return self.__class__.__name__
        return f"{self.__class__.__name__} {signature(self.func)}"

    def __repr__(self):
        return self.__str__()

    @classmethod
    def from_function(cls, func: Callable, *init_args, **init_kwargs):
        # TODO: some of this boilerplate could be reduced with a bunch of
        #  fancy getattr overrides
        dynamic = super().__new__(cls)
        try:
            dynamic.source = getsource(func)
        except TypeError:  # it's some kind of weirdo callable
            # noinspection PyUnresolvedReferences
            func = func.__call__
            dynamic.source = getsource(func)
        dynamic.code = func.__code__
        dynamic.func = func
        dynamic.__signature__ = signature(dynamic.func)
        dynamic.__name__ = dynamic.func.__name__
        dynamic.globals_ = func.__globals__
        dynamic.__init__(*init_args, **init_kwargs)
        return dynamic

    __signature__ = Signature()
    source, code, func, __name__ = None, None, None, '<unloaded Dynamic>'



import datetime as dt
import json
from inspect import currentframe, getframeinfo
from io import SEEK_END
from itertools import count
import os
from pathlib import Path


class TrivialTracker:
    def dump(self):
        pass

    def track(self, *_, **__):
        pass

    def clear(self):
        pass

    def set_metadata(self, **metadata):
        pass

    log = None
    history = None


class Tracker(TrivialTracker):
    """watches where it's been"""

    def __init__(self, name=None, identifier="", outdir=None):
        self.history, self.metadata, self.counter = [], {}, count(1)
        self.name = name
        self.log = {
            "name": self.name,
            "init_timestamp": dt.datetime.now().isoformat(),
            "history": self.history,
        }
        if outdir is None:
            outdir = Path(__file__).parent.parent / ".tracker_logs"
        if identifier != "":
            identifier = f"{identifier}_"
        outdir.mkdir(exist_ok=True)
        self.outpath = Path(outdir, f"{identifier}{self.name}.json")

    def clear(self):
        self.outpath.unlink(missing_ok=True)
        self.history[:] = []
        self.counter = count(1)

    def track(self, func, **metadata):
        if "__name__" in dir(func):
            target = func.__name__
        else:
            target = str(func)
        caller = currentframe().f_back
        info = getframeinfo(caller)
        rec = (
            {
                "target": target,
                "caller": info.function,
                "lineno": info.lineno,
                "trackcount": next(self.counter),
            }
            | self.metadata
            | metadata
        )
        self.history.append(rec)

    def set_metadata(self, **metadata):
        self.metadata |= metadata

    def _get_paused(self):
        return self._paused

    def _set_paused(self, state: bool):
        self._paused = state

    # noinspection PyTypeChecker
    def dump(self):
        if self.paused is True:
            return
        self.log["write_timestamp"] = dt.datetime.now().isoformat()
        key_order = sorted(self.log.keys(), key=lambda n: "history" in n)
        self.log = {k: self.log[k] for k in key_order}
        mode = "rb+" if self.outpath.exists() else "wb"
        with self.outpath.open(mode) as stream:
            if mode == "rb+":
                stream.seek(-1, SEEK_END)
                stream.write(b",\n")
            else:
                stream.write(b"[")
            stream.write(json.dumps(self.log, indent=2).encode('utf-8'))
            stream.write(b"]")

    _paused = False
    paused = property(_get_paused, _set_paused)
